#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <time.h>
#include <stdlib.h>
#include <sys/file.h>
#include "svstat.h"
#include "iolib.h"
const char *progname = "supervise";
#include "freedt.h"

struct svstat svstatus;
int die_with_child = 0;
int restart_on_death = 1;
int statusfd, control;
int sigchld_received = 0;

void sigchld_handler(int value) {
	sigchld_received = 1;
}

void start_child() {
	if (svstatus.pid != -1) return;

	svstatus.pid = fork();
	if (svstatus.pid < 0)
		die("fork failed");
	else if (svstatus.pid == 0) {
		close(control);
		close(statusfd);

		if (setsid() < 0)
			_exit(111);
		execl("./run", "run", NULL);
		_exit(111);
	}
	svstatus.up = 1;
	svstatus.starttime = time(NULL);

	while (sleep(1) != 0)
		;
}

void kill_child(int signal) {
	if (svstatus.pid == -1) return;

	if (kill(svstatus.pid, signal) < 0)
		die("unable to send signal to process");
}

void write_status() {
	if (lseek(statusfd, 0, SEEK_SET) < 0)
		die("unable to seek status file");
	if (write(statusfd, &svstatus, sizeof svstatus) < 0)
		die("unable to write to status file");
}

int main(int argc, char **argv) {
	struct sigaction sa;
	int fd;

	istdinit();

	svstatus.pid = -1;
	svstatus.starttime = 0;
	svstatus.up = 0;

	if (argc < 1)
		die("usage: supervise service");

	if (chdir(argv[1]) < 0)
		die2(argv[1], "unable to chdir to service directory");

	if (mkdir("supervise", 02700) < 0 && errno != EEXIST)
		die2(argv[1], "unable to create supervise directory");

	fd = open("supervise/lock", O_WRONLY | O_CREAT, 0600);
	if (fd < 0)
		die2(argv[1], "unable to open supervise/lock");
	if (flock(fd, LOCK_EX) < 0)
		die2(argv[1], "unable to obtain lock on supervise/lock");

	if (mkfifo("supervise/control", 0600) < 0 
		&& errno != EEXIST)
		die2(argv[1], "unable to create fifo supervise/control");
	control = open("supervise/control", O_RDWR | O_NONBLOCK);
	if (control < 0)
		die2(argv[1], "unable to open fifo supervise/control");

	statusfd = open("supervise/status", O_WRONLY | O_CREAT, 0644);
	if (statusfd < 0)
		die2(argv[1], "unable to open supervise/status");

	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGCHLD, &sa, NULL) < 0)
		die("unable to install SIGCHLD handler");

	fd = open("down", O_RDONLY);
	if (fd < 0) {
		start_child();
	} else {
		svstatus.starttime = time(NULL);
		close(fd);
	}

	write_status();

	while (1) {
		char c;
		int rc;

		do {
			if (sigchld_received) {
				int wrc;

				do {
					/* The child's status has changed. */
					wrc = waitpid(svstatus.pid, NULL,
						WNOHANG);
					if (wrc < 0)
						die("waitpid failed");
	
					if (wrc > 0) {
						svstatus.pid = -1;
						if (die_with_child)
							exit(0);
						if (restart_on_death)
							start_child();
						else
							svstatus.up = 0;
					}
				} while (wrc > 0);

				svstatus.starttime = time(NULL);
				write_status();
				sigchld_received = 0;
			}

			rc = read(control, &c, 1);
		} while (rc == 0 || (rc < 0 && errno == EINTR));
		if (rc < 0)
			die("read from control failed");

		/* Command from svc. */

		switch (c) {
		case 'u':
			restart_on_death = 1;
			start_child();
			write_status();
			break;
		case 'o':
			restart_on_death = 0;
			start_child();
			write_status();
			break;
		case 'd':
			restart_on_death = 0;
			kill_child(SIGTERM);
			kill_child(SIGCONT);
			break;
		case 'p':
			kill_child(SIGSTOP);
			break;
		case 'c':
			kill_child(SIGCONT);
			break;
		case 'h':
			kill_child(SIGHUP);
			break;
		case 'a':
			kill_child(SIGALRM);
			break;
		case 'i':
			kill_child(SIGINT);
			break;
		case 't':
			kill_child(SIGTERM);
			break;
		case 'k':
			kill_child(SIGKILL);
			break;
		case 'X':
			kill_child(SIGTERM);
			kill_child(SIGCONT);
			exit(0);
		case 'x':
			if (svstatus.up == 1)
				die_with_child = 1;
			else
				exit(0);
			break;
		}
	}

	return 0; /* NOTREACHED */
}

